/******************************************************************

        MUSCLE SmartCard Development ( http://www.linuxnet.com )
            Title  : cardman.c
            Package: Utimaco Serial Driver
            Author : David Corcoran
            Date   : 3/26/00
            Purpose: This provides Cardman specific functionality.

********************************************************************/   

#include <pcscdefines.h>
#include <LinuxDefines.h>
#include <AdmHndlr.h>
#include <cardman.h>
#include <serial.h>

#define CARDMODE    1
#define READERMODE  2
#define CARDNO      0
#define CARDT0      1
#define CARDT1      2

struct _CardManOpts {
  int cmanmode;
  int cardmode;
} CardManOpts;

	
void CMAN_DebugLog( char *pcMessage, char *pcFile, int iLine ) {
  printf("%25s:%05d: %s\n", pcFile, iLine, pcMessage);
}

int CMAN_Command( unsigned char ucConfig[], unsigned char ucHeader[],
		  unsigned char ucControl ) {

  int i, status;

  i=0; status=0;

  status = IO_Write( ucConfig, 4 );
  if (status == FALSE) {
    CMAN_DebugLog( "CMAN_Command: Error\n", __FILE__, __LINE__ ); 
    return STATUS_IO_TIMEOUT;
  }    

  IO_Write( ucHeader, 4 );
  if (status == FALSE) {
    CMAN_DebugLog( "CMAN_Reset: ReSync Error\n", __FILE__, __LINE__ );
    return STATUS_IO_TIMEOUT;
  }      

  status = IO_Write( &ucControl, 1 );

  if ( status == FALSE ) {
    CMAN_DebugLog( "CMAN_Reset: ReSync Error\n", __FILE__, __LINE__ );
    return STATUS_IO_TIMEOUT;
  }

  return STATUS_SUCCESS;
}

int CMAN_Initialize( char *pcPortPath ) {

  int status;

  status = IO_InitializePort (9600, 8, 'E', pcPortPath);
  
  if ( status == TRUE ) {
    status = CMAN_ResynchReader();
    return status;
  } else {
    return STATUS_UNSUCCESSFUL;
  }
}

int CMAN_UnInitialize() {

  int status;

  status = IO_Close();

  if ( status == TRUE ) {
    return STATUS_SUCCESS;
  } else {
    return STATUS_UNSUCCESSFUL;
  }
}

int CMAN_ResynchFE( ) {

  unsigned char buffFE[CARDMAN_RESYNCH_NUM];
  unsigned char buff00 = 0x00;
  int i, status;

  i=0; status=0;

  memset( buffFE, 0xFE, CARDMAN_RESYNCH_NUM );
  
  status = IO_Write ( buffFE, CARDMAN_RESYNCH_NUM );
  if (status == FALSE) {
    CMAN_DebugLog( "ReSyncFE Error\n", __FILE__, __LINE__ );
    return STATUS_IO_TIMEOUT;
  }
  
  status = IO_Write( &buff00, 1 );
  
  if ( status == FALSE ) {
    CMAN_DebugLog( "Sync Error\n", __FILE__, __LINE__ );
    return STATUS_IO_TIMEOUT;
  }

  return STATUS_SUCCESS;
}

int CMAN_Configure( ) {

  unsigned char buff89[5] = {0x00, 0x00, 0x00, 0x00, 0x89};
  int i, status;

  i=0; status=0;

  status = IO_Write ( buff89, 5 );
  if (status == FALSE) {
    CMAN_DebugLog( "ReSync89 Error\n", __FILE__, __LINE__ );
    return STATUS_IO_TIMEOUT;
  }

  return STATUS_SUCCESS;
}


int CMAN_ResynchReader( ) {

  int i, status;
  unsigned char stat;
  unsigned char ctlend = 0x89;
  unsigned char config[4] = { 0x03, 0x00, 0x00, 0x80 };
  
  i=0; status=0; stat=0;


  /* Resynch FE   */
 
  status = CMAN_ResynchFE();

  if ( status != STATUS_SUCCESS ) {
    return status;
  }


  /* Write Config */
  
  status = IO_Write( config, 4 );
  if (status == FALSE) {
    CMAN_DebugLog( "Sync Error\n", __FILE__, __LINE__ );
    return STATUS_IO_TIMEOUT;
  }    

  CMAN_Status( &stat );

  status = IO_Write( &ctlend, 1 );    
  if (status == FALSE) {
    CMAN_DebugLog( "Sync Error\n", __FILE__, __LINE__ );
    return STATUS_IO_TIMEOUT;
  }    
  
  CardManOpts.cmanmode = READERMODE;

  return STATUS_SUCCESS;
}

/* Puts the reader in transparent mode so that all commands 
   are transmited directly to the card and same on the return. */

int CMAN_ResynchCard( ) {

  int i, status;
  unsigned char stat;
  const unsigned char buff00 = 0x00;
  unsigned char config[4] = { 0x97, 0x00, 0x00, 0x80 };
  unsigned char header[4] = { 0x00, 0x00, 0x00, 0x00 };

  i=0; status=0; stat=0;

  /* Resynch FE   */
 
  status = CMAN_ResynchFE();

  if ( status != STATUS_SUCCESS ) {
    return status;
  }
  
  status = CMAN_Command( config, header, buff00 );
  if ( status != STATUS_SUCCESS ) {
    return status;
  }
  
  CardManOpts.cmanmode = CARDMODE;
  return STATUS_SUCCESS;
}

int CMAN_Status( unsigned char *stat ) {

  int i, status;
  unsigned char header[4]    = { 0x00, 0x00, 0x00, 0x00 };
  unsigned char buff00 = 0x00;
  unsigned char stabuf[4];

  i=0; status=0;
  memset( stabuf, 0x00, sizeof( stabuf ));
    
  status = IO_Write( header, 4 );
  if (status == FALSE) {
    CMAN_DebugLog( "CMAN_Status: Header Error\n", __FILE__, __LINE__ );
    return STATUS_IO_TIMEOUT;
  }    

  status = IO_Write( &buff00, 1 );
  
  if ( status == FALSE ) {
    CMAN_DebugLog( "CMAN_Status: 00 Error\n", __FILE__, __LINE__ );
    return STATUS_IO_TIMEOUT;
  }
  

  if (!IO_Read(2, stabuf)) {
    CMAN_DebugLog( "CMAN_Status: Read Error\n", __FILE__, __LINE__ );
    return STATUS_IO_TIMEOUT;
  }

  /* Gobble up that insertion byte */
  if ( stabuf[0] == 0x00 ) {
    stabuf[0] = stabuf[1];
    if (!IO_Read(1, &stabuf[1])) {
      CMAN_DebugLog( "CMAN_Status: Read Error\n", __FILE__, __LINE__ );
      return STATUS_IO_TIMEOUT;
    }
  }

  *stat = stabuf[0]; 

  if( stabuf[0] == 0x62 ) { /* Card not inserted */
      return STATUS_SUCCESS;
    } else {
      if( stabuf[0] == 0x66 || stabuf[0] == 0x64)/* 66=card in, no power */
	{                                        /* 64=card in & power   */
	  return STATUS_SUCCESS;
	} else {
	  printf("Buf is %x\n", stabuf[0]);
	  CMAN_DebugLog( "CMAN_Status: Invalid Error\n", __FILE__, 
			 __LINE__ );
	  return STATUS_IO_TIMEOUT;	  
	}
    }
  
}

int CMAN_Reset( unsigned char *atr, int *atrsize ) {

  int i, status, atrlen;
  unsigned char stat;
  unsigned char stabuf[40];
  const unsigned char ctlchr = 0x00;
  const unsigned char buff00 = 0x00;
  unsigned char config[4]    = { 0x87, 0x00, 0x02, 0x02 };
  unsigned char configdba[4] = { 0x83, 0x00, 0x00, 0x80 };
  unsigned char configdbb[4] = { 0x93, 0x00, 0x00, 0x80 };
  unsigned char header[4]    = { 0x00, 0x00, 0x40, 0x21 };

  i=0; status=0; stat=0; atrlen=0;

  CMAN_DebugLog("CMAN_Reset: Resetting Card\n", __FILE__, __LINE__ );

  if ( CardManOpts.cmanmode == CARDMODE ) {
    IO_CardManpulserts ();
  }

  /* Resynch FE   */
 
  status = CMAN_ResynchFE();
  if ( status != STATUS_SUCCESS ) {
    return status;
  }

  status = CMAN_Command( config, header, buff00 );
  if ( status != STATUS_SUCCESS ) {
    return status;
  }
  
  if (!IO_Read(2, stabuf)) {
    CMAN_DebugLog( "CMAN_Reset: ReSync Error\n", __FILE__, __LINE__ );
    return STATUS_IO_TIMEOUT;
  }

  if ( stabuf[0] == 0x66 || stabuf[0] == 0x64 ) {
    atrlen = stabuf[1];
  }

  if ( atrlen == 0 ) {
    /* Have to use Dump Buffer strategy */

    /* Resynch FE   */
    
    status = CMAN_ResynchFE();
    
    if ( status != STATUS_SUCCESS ) {
      return status;
    }
   
    status = CMAN_Command( configdba, header, buff00 );
    if ( status != STATUS_SUCCESS ) {
      return status;
    }
    
    sleep(2);
    
    /* Resynch FE   */
    
    status = CMAN_ResynchFE();
    
    if ( status != STATUS_SUCCESS ) {
      return status;
    }
    
    status = CMAN_Command( configdbb, header, buff00 );
    if ( status != STATUS_SUCCESS ) {
      return status;
    }

    if (!IO_Read(2, stabuf)) {
      CMAN_DebugLog( "CMAN_Reset: ReSync Error\n", __FILE__, __LINE__ );
      return STATUS_IO_TIMEOUT;
    }

    if ( stabuf[0] == 0x66 || stabuf[0] == 0x64 ) {
      atrlen = stabuf[1];
    }

    if ( atrlen == 0 ) {
      *atrsize = 0;
      return STATUS_IO_TIMEOUT;
    }

  }  /* End of Dump Buffer Strategy */
  
  if (!IO_Read(atrlen, atr)) {
    return STATUS_IO_TIMEOUT;
  } else {
    *atrsize = atrlen;
    return STATUS_SUCCESS;
  }

}

int CMAN_IsCardInserted( ) {

  int status, failtry;
  unsigned char stat;

  failtry = 0;
  status  = 0;
  stat    = 0;

  if ( CardManOpts.cmanmode == CARDMODE ) {
    IO_CardManpulserts ();
    CMAN_ResynchReader();
  }

  status = CMAN_Status( &stat );
  
  if ( status != STATUS_SUCCESS ) {
    CMAN_DebugLog( "CMAN_IsCardInserted: Status Error\n", __FILE__, __LINE__ );
    return status;
  }
  
  if( stat == 0x66 || stat == 0x64 ) {   
    return STATUS_SUCCESS;      
  } else if ( stat == 0x62 ) {
    return STATUS_UNSUCCESSFUL;
  } else {
    return STATUS_IO_TIMEOUT;
  }

}
int CMAN_IsCardPowered( ) {

  int status;
  unsigned char stat;
  
  status = 0;
  stat   = 0;

  status = CMAN_Status( &stat );

  if ( status != STATUS_SUCCESS ) {
    CMAN_DebugLog( "CMAN_IsCardPowered: Status Error\n", __FILE__, __LINE__ );
    return status;
  }

  
  if( stat == 0x64 ) {   
    return STATUS_SUCCESS;      
  } else {
    return STATUS_UNSUCCESSFUL;
  }

}

int CMAN_TransmitT0( unsigned int RequestLen, unsigned char* pRequest, 
		     unsigned int* pReplyLen, unsigned char *pReply ) {

  int i, len, rv;
  int commtype = 0;
  int bytesWritten, bytesToWrite;
  int totalBytesToWrite;
  int isoApduCase; 
  unsigned long rspSize, ulSlot;
  unsigned char rxBuffer[MAX_BUFFER_SIZE];
  unsigned char rsp[MAX_BUFFER_SIZE];
  unsigned char cmd[MAX_BUFFER_SIZE];

  ulSlot = 1;

  Adm_SetWWT(ADM_DEFAULT_WWT);

  if ( CardManOpts.cmanmode == READERMODE ) {
    /* Switch to transparent mode */
    CMAN_ResynchCard();
  }

  /* Determine the type of APDU to be sent */

  if ( (RequestLen == 5) && (pRequest[4] == 0x00) ) {
    isoApduCase = 1;  /* ISO 7816 Case 1 Transaction */
  } else if ( (RequestLen == 5) && (pRequest[4] > 0x00) ) {
    isoApduCase = 2;  /* ISO 7816 Case 2 Transaction */
  } else if ( (RequestLen > 5) && (pRequest[4] > 0x00) ) {
    isoApduCase = 3;  /* ISO 7816 Case 3 Transaction */
  } else if ( (RequestLen > 5) && ((RequestLen) == 
              (pRequest[4] + 5 + 1)) ) {
    isoApduCase = 4;  /* ISO 7816 Case 3 Transaction */
  }

  bytesWritten      = 0;
  totalBytesToWrite = RequestLen;

  bytesToWrite = 5;

  do {

    rspSize = 1;  /* Only get 1 byte from the card */

    memcpy( cmd, &pRequest[bytesWritten], bytesToWrite );

    rv = Adm_TransmitICC( ulSlot, cmd, bytesToWrite, 
			  rsp, &rspSize );
    
    if ( rv != STATUS_SUCCESS ) {
      return rv;
    }

    bytesWritten      += bytesToWrite;
    totalBytesToWrite -= bytesToWrite;

    if (rsp[0] == 0x60) {
      usleep(50000);    /* Sleep for a bit     */
      bytesToWrite = 0; /* Just read this time */
      continue;
      
    } else if ( rsp[0] == cmd[1] ) {
      if ( isoApduCase == 1 ) {
        rspSize = 2;
        rv = Adm_TransmitICC( ulSlot, cmd, 0, 
			      rsp, &rspSize );
        *pReplyLen = 2;
        memcpy( pReply, rsp, rspSize );
        return rv;

      } else if ( isoApduCase == 2 ) {
        rspSize = pRequest[4] + 2;
        rv = Adm_TransmitICC( ulSlot, cmd, 0, 
			      rsp, &rspSize );
        *pReplyLen = rspSize;
        memcpy( pReply, rsp, rspSize );
        return rv;

      } else if ( isoApduCase == 3 ) {
        rspSize = 2;
        printf("Length to send is %d\n", totalBytesToWrite);
        memcpy( cmd, &pRequest[bytesWritten], totalBytesToWrite );
        rv = Adm_TransmitICC( ulSlot, cmd, totalBytesToWrite, 
			      rsp, &rspSize );
        *pReplyLen = 2;
        memcpy( pReply, rsp, rspSize );
        return rv;
      } else if ( isoApduCase == 4 ) {
        /* Stupid ass ISO 7816 brain dead transaction */
      } else {
        /* We have a problem here.... */
        return STATUS_UNSUCCESSFUL;
      }

   } else if ( (rsp[0] == ~cmd[1]) || 
               (rsp[0] == ~(cmd[1] + 1)) ) {
      
      if ( isoApduCase == 1 ) {
         /* This will never happen */
         return STATUS_UNSUCCESSFUL;
       } else if ( isoApduCase == 2 ) {
         /* we can only receive the next byte */
         bytesToWrite = 0;
         continue;
       } else if ( isoApduCase == 3 ) {
         /* we can only send the next byte */
         bytesToWrite = 1;
         continue;
       } else if ( isoApduCase == 4 ) {
        /* ISO 7816 case 4 transaction */
       }

   } else if (( rsp[0] >= 0x61 && rsp[0] <= 0x6F ) ||
             ( rsp[0] >= 0x90 && rsp[0] <= 0x9F )) {
     
     pReply[0] = rsp[0];

     rspSize = 1;
     rv = Adm_TransmitICC( ulSlot, cmd, 0, 
			   rsp, &rspSize );
     *pReplyLen = 2;
     pReply[1] = rsp[0];

     return rv;
   } else {
     /* Serious problem here */
   }

  } while (1);


  return STATUS_SUCCESS;
}


int CMAN_TransmitT1( unsigned int lc, unsigned char* cmd, 
		     unsigned int* lr, unsigned char *rsp ) {

  int i, len, status, retsize;
  int commtype = 0;
  unsigned char buff00 = 0x00;
  unsigned char config[] = {0x82, 0x40, 0x00, 0x80};
  unsigned char header[] = {0x00, 0x00, 0x00, 0x03};
  unsigned char txBuffer[MAX_BUFFER_SIZE];
  unsigned char rxBuffer[MAX_BUFFER_SIZE];

  header[1] = lc;

  status = IO_Write( config, 4 );
  if (status == FALSE) {
    CMAN_DebugLog( "CMAN_Command: Error\n", __FILE__, __LINE__ ); 
    return STATUS_IO_TIMEOUT;
  }    

  IO_Write( header, 4 );
  if (status == FALSE) {
    CMAN_DebugLog( "CMAN_Reset: ReSync Error\n", __FILE__, __LINE__ );
    return STATUS_IO_TIMEOUT;
  }      

  status = IO_Write( cmd, lc );  
  if ( status == FALSE ) {
    CMAN_DebugLog( "CMAN_TransmitT1: Error\n", __FILE__, __LINE__ );
    return STATUS_IO_TIMEOUT;
  } 

  status = IO_Write( &buff00, 1 );
  if ( status == FALSE ) {
    CMAN_DebugLog( "CMAN_TransmitT1: Error\n", __FILE__, __LINE__ );
    return STATUS_IO_TIMEOUT;
  } 

  status = IO_Read( 2, rxBuffer );
  if ( status == FALSE ) {
    CMAN_DebugLog( "CMAN_TransmitT1: Error\n", __FILE__, __LINE__ );
    return STATUS_IO_TIMEOUT;
  } 

  if ( rxBuffer[0] == 0x64 ) {
    retsize = rxBuffer[1];
  } else {
    CMAN_DebugLog( "CMAN_TransmitT1: Error\n", __FILE__, __LINE__ );
    return STATUS_IO_TIMEOUT;
  }

  status = IO_Read( retsize, rsp );
  if ( status == FALSE ) {
    *lr = retsize;    
    CMAN_DebugLog( "CMAN_TransmitT1: Error\n", __FILE__, __LINE__ );
    return STATUS_IO_TIMEOUT;
  } 
    
  return STATUS_SUCCESS;
}


